/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Mobile Application Link.
 *
 * The Initial Developer of the Original Code is AvantGo, Inc.
 * Portions created by AvantGo, Inc. are Copyright (C) 1997-1999
 * AvantGo, Inc. All Rights Reserved.
 *
 * Contributors: Jason Day <jasonday@worldnet.att.net>
 *               Reinhold Kainhofer (reinhold@kainhofer.com)
 *
 */
/*

  See the README file in this directory for info on compiling, using
  and the state of this program. If the readme file is for some reason
  not there, it can be found at the MAL website:

  http://corp.avantgo.com/mal/

*/
/* Thu, 29 Jul 1999 12:43:01 +0200
 *     Turbo Fredriksson <turbo@nocrew.org>
 *     Added daemon mode (close STDIO and disconnect from the shell
 *     Rewrote the command line parsing (to use getopt_long)
 */
/*
 * 25 Jan 2000
 * Jason Day <jasonday@worldnet.att.net>
 * Added code for submission of .sub files
 */
/*
 * 15 Mar 2000
 * Jason Day <jasonday@worldnet.att.net>
 * Put most of the code into a library
 */

#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <dlfcn.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include <AGNet.h>
#include <AGUtil.h>
#include <AGClientProcessor.h>
#include <AGProtocol.h>
#include <AGBufferReader.h>
#include <AGPalmProtocol.h>
#include <AGUserConfig.h>
#include <AGServerConfig.h>
#include <AGSyncCommon.h>
#include <AGCommandProcessor.h>
#include <AGDesktopInfoPalm.h>
#include <AGTimeoutMessageBox.h>
#include <AGMobileLinkSettings.h>
#include <AGMD5.h>
#include <MAL31UserConfig.h>


#include <pi-source.h>
#include <pi-socket.h>
#include <pi-file.h>
#include <pi-dlp.h>
#include <pi-version.h>

#include "config.h"
#include "libmal.h"

#define DEFAULT_CARD_NUM 0

static char *device = "/dev/pilot";
static char *httpProxy      = NULL;
static int   httpProxyPort  = 0;
static char *socksProxy     = NULL;
static int   socksProxyPort = 0;
static char *proxyUsername  = NULL;
static char *proxyPassword  = NULL;
static char *subfile        = NULL;
static int threeone = 0;

#ifdef _DEBUG
#define debug_print(x) printf("%s\n", (x));
#else
#define debug_print(x)
#endif

/* Pilot-link version */
#if PILOT_LINK_VERSION <= 0
# if PILOT_LINK_MAJOR <= 9
#  define PILOT_LINK_0_9
# endif
#endif
#ifdef AP_SNPRINTF
#define snprintf ap_snprintf
#define vsnprintf ap_vsnprintf
#endif

int null_status(const char *null, ...) {
    return 0;
}
static printStatusHook printstatus = printf;
static printErrorHook printerror = printf;


static netInitFunc     secnetinit = NULL;
static netCloseFunc    secnetclose = NULL;
static netCtxSizeFunc  secctxsize = NULL;
static netPreSyncHook  secnetpresync = NULL;
static netPostSyncHook  secnetpostsync = NULL;
static int sd = 0;
static int verbose = 1;


/* static function prototypes */
static int32 initAndOpenDatabase(void *out, AGDBConfig *theDatabase, 
                                 int32 *errCode);
static int32 getNextModifiedRecord(void * out, AGRecord ** record, 
                                   int32 *errCode);
static int32 getNextRecord(void * out, AGRecord ** record,
                           int32 *errCode);

static int32 cmdTASK(void *out, int32 *returnErrorCode, char *currentTask,
                     AGBool bufferable);
static int32 cmdITEM(void *out, int32 *returnErrorCode,
                                   int32 currentItemNumber,
                                   int32 totalItemCount,
                                   char *currentItem);
static int32 cmdDELETEDATABASE(void *out, int32 *returnErrorCode,
                                             char *dbname);
static int32 cmdOPENDATABASE(void *out, int32 *returnErrorCode,
                                           char *dbname);
static int32 cmdCLOSEDATABASE(void *out, int32 *returnErrorCode);
static int32 cmdCLEARMODS(void *out, int32 *returnErrorCode);
static int32 cmdGOODBYE(void *out, int32 *returnErrorCode,
                        AGSyncStatus syncStatus,
                        int32 errorCode,
                        char *errorMessage);
static int32 cmdRECORD(void *out, int32 *returnErrorCode,
                       int32 *newUID,
                       int32 uid,
                       AGRecordStatus mod,
                       int32 recordDataLength,
                       void *recordData,
                       int32 platformDataLength,
                       void *platformData);

static RETSIGTYPE SigHandler(int signal);
static long openUserConfigDatabase(int *threeone);
static int32 readDeviceUserConfig32(int userConfigDBHandle, AGUserConfig **deviceUserConfig);
static int32 readDeviceUserConfig31(int userConfigDBHandle, AGUserConfig **deviceUserConfig);
static int32 readDeviceUserConfig(int userConfigDBHandle,
                                  AGUserConfig **deviceUserConfig,
                                  int threeone);
static void writeDeviceUserConfig(int userConfigDBHandle,
                                  AGUserConfig * deviceUserConfig,
                                  recordid_t *recID, int threeone);

/* ----------------------------------------------------------------------------*/
/* defines and prototypes for the .sub stuff */
#define kAvantGoSubXMLWrapper "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\
<!DOCTYPE avantgo_subs_file [ \
<!ELEMENT sub (#PCDATA)>\
]>\
<sub>%s</sub>"
#define kSubmissionURI "/sync/sub"
#define kSubSubmissionGoodMessage "Subs file submitted successfully"
#define kSubmissionGoodMessage "File submitted successfully"

static int32 getNextCommand(void *out,
                            int32 *newCommand,
                            int32 *commandLength,
                            void **commandBytes);
static char *openSubFile (char **buf, int *bufsize);

/* ----------------------------------------------------------------------------
 * These two functions are a hack to allow callers to update a GUI with
 * the current status of the sync. Callers may register a callback function
 * which will be then called by the MAL commands when something happens.
 */
void register_printStatusHook (printStatusHook pfn) {
    printstatus = pfn;
}

void register_printErrorHook (printErrorHook pfn) {
    printerror = pfn;
}
 

/*
 * Setter methods for the proxy and socks variables.
 */

void setHttpProxy (char *str) {
    httpProxy = str;
}

void setHttpProxyPort (int port) {
    httpProxyPort = port;
}

void setSocksProxy (char *str) {
    socksProxy = str;
}

void setSocksProxyPort (int port) {
    socksProxyPort = port;
}

void setProxyUsername (char *str) {
    proxyUsername = str;
}

void setProxyPassword (char *str) {
    proxyPassword = str;
}

/* ----------------------------------------------------------------------------*/
void syncInfoFree(PalmSyncInfo * pInfo)
{

    if (NULL != pInfo) {

        if (NULL != pInfo->platform)
            free(pInfo->platform);

        if (NULL != pInfo->userConfig)
            AGUserConfigFree(pInfo->userConfig);

        if (NULL != pInfo->pilot_buffer)
            free(pInfo->pilot_buffer);

        if (NULL != pInfo->commandProcessor)
            AGCommandProcessorFree(pInfo->commandProcessor);

        free(pInfo);
    }
}

/* ----------------------------------------------------------------------------*/
PalmSyncInfo * syncInfoNew()
{
    PalmSyncInfo * pInfo;

    /* Allocate the PalmSyncInfo structure. */
    pInfo = (PalmSyncInfo *)malloc(sizeof(PalmSyncInfo));
    if (NULL != pInfo) {

        const int pbs = 65535;

        bzero(pInfo, sizeof(PalmSyncInfo));

        pInfo->pilot_buffer_size    = pbs;
        pInfo->pilot_buffer         = (uint8 *)malloc(pbs);
        if (NULL == pInfo->pilot_buffer)
            goto fail;

        /* Allocate the platform calls record. */
        pInfo->platform = (AGPlatformCalls *)malloc(sizeof(AGPlatformCalls));
        bzero(pInfo->platform, sizeof(AGPlatformCalls));
        if (NULL == pInfo->platform)
            goto fail;

        return pInfo;
    }

fail:
    if (NULL != pInfo)
        syncInfoFree(pInfo);

    return NULL;
}
/* ----------------------------------------------------------------------------*/
static AGBool setupPlatformCalls(PalmSyncInfo * pInfo)
{
    pInfo->platform->out = pInfo;
    pInfo->platform->nextModifiedRecordFunc = getNextModifiedRecord;
    pInfo->platform->nextRecordFunc = getNextRecord;
    pInfo->platform->openDatabaseFunc =  initAndOpenDatabase;
    return FALSE;
}
/* ----------------------------------------------------------------------------*/
static int readDeviceInfo(PalmSyncInfo * pInfo)
{
    AGDeviceInfo * devInfo;
    struct SysInfo sysInfo;
    struct CardInfo cardInfo;
    int majorver, minorver;
    char osverstring[24];
    int err;
    
    /* Start with clean slate. */
    devInfo = pInfo->deviceInfo;

    err = dlp_ReadSysInfo(sd, &sysInfo);
    if (err < 0) {
	(*printerror) ("dlp_ReadSysInfo failed with err %d\n", err);
        return -1;
    }

    cardInfo.card = DEFAULT_CARD_NUM;
    err = dlp_ReadStorageInfo(sd, DEFAULT_CARD_NUM, &cardInfo);
    if (err < 0) {
        /*fprintf(stderr, "dlp_ReadStorageInfo failed with err %d\n", err);*/
        (*printerror) ("dlp_ReadStorageInfo failed with err %d\n", err);
        return -1;
    }

    /* Find out version of operating system. */
    majorver = sysInfo.romVersion >> 24;
    minorver = 0xff & (sysInfo.romVersion >> 16);
    if (minorver < 10) 
        minorver = 0;
    
    snprintf(osverstring, 24, "%d.%d", majorver, minorver);

    devInfo->availableBytes   = cardInfo.ramFree;
    devInfo->serialNumber     = strdup("");
    devInfo->osName           = strdup("PALM_OS");
    devInfo->osVersion        = strdup(osverstring);
    devInfo->screenWidth      = 150;
    devInfo->screenHeight     = 150;
    if(majorver > 2)
       devInfo->colorDepth    = 2; 
    else
       devInfo->colorDepth    = 1; 

    return 0;
}
/* ----------------------------------------------------------------------------*/
static AGBool getPalmDatabaseCreationInfo(AGDBConfig * db, uint32 * creator,
                                          uint32 * flags, uint32 * type)
{
    AGBufferReader * r = NULL;
    debug_print("GetPalmDatabaseCreationInfo()\n");

    if (NULL == db)
        return FALSE;

    if (0 == db->platformDataLength || NULL == db->platformData)
        return FALSE;

    r = AGBufferReaderNew((uint8*)db->platformData);
    if (NULL == r)
        return FALSE;
    
    AGPalmReadDBConfigPlatformData((AGReader*)r, creator, type, flags);
    AGBufferReaderFree(r);
    return TRUE;
}

/* ----------------------------------------------------------------------------*/
static int createDatabase(AGDBConfig *db)
{
    int dbhandle;
    long creator;
    int flags;
    int cardNo = DEFAULT_CARD_NUM;
    long type;
    int version = 0;
    int err;
    
    if (!db)
        return 0;

#ifdef _DEBUG
    printf("createDatabase\n");
#endif
    
    getPalmDatabaseCreationInfo(db, (uint32*)&creator, 
                                (uint32*)&flags, (uint32*)&type);

    if ((err = dlp_CreateDB(sd,  creator,  type, cardNo, flags,  
                            version, db->dbname, &dbhandle)) < 0) {
#ifdef _DEBUG
        printf("createDatabase: dlp_CreateDB failed err = %d\n", err);
#endif
        
        dbhandle = 0;
    }
    
    return dbhandle;
}

/* ----------------------------------------------------------------------------*/
static long openDatabase(PalmSyncInfo *pInfo, char *dbname, AGBool create)
{

    long result;
    
    if (!dbname || !pInfo) {
        debug_print("\n");
        return -1;
    }
    debug_print("... opening '");
    debug_print(dbname);
    debug_print("' ... ");

    pInfo->currentDb =
        AGServerConfigGetDBConfigNamed(pInfo->currentServerConfig, dbname);
    
    result = dlp_OpenDB(sd, DEFAULT_CARD_NUM, dlpOpenRead|dlpOpenWrite,  
                        dbname, &pInfo->pilot_rHandle);
    
    if ((result < 0) && create)
        pInfo->pilot_rHandle = createDatabase(pInfo->currentDb);

    if (pInfo->pilot_rHandle) {
        uint32 creator, flags, type;
        if (getPalmDatabaseCreationInfo(pInfo->currentDb, &creator, &flags, 
                                        &type) && (flags & 0x01) != 0) {
            pInfo->currentDBIsResourceType = TRUE;
        } else
            pInfo->currentDBIsResourceType = FALSE;
        debug_print("successfully.\n");
    } else {
        debug_print("unsuccessfully.\n");
        pInfo->currentDBIsResourceType = FALSE;
        pInfo->currentDb = NULL;
    }

    return result;
}

/* ----------------------------------------------------------------------------*/
static long closeDatabase(PalmSyncInfo * pInfo)
{
    int result;

    result = dlp_CloseDB(sd, pInfo->pilot_rHandle);
    pInfo->pilot_rHandle = 0;
    pInfo->currentDb = NULL;
    pInfo->currentDBIsResourceType = FALSE;
    return result;
}

/* ----------------------------------------------------------------------------*/
static void deleteDatabase(char * dbname)
{
    debug_print("deleteDatabase()\n");
    dlp_DeleteDB(sd, DEFAULT_CARD_NUM, dbname);
}
/* ----------------------------------------------------------------------------*/
static void clearMods(long dbHandle)
{
    debug_print("clearMods()\n");
    dlp_CleanUpDatabase(sd, dbHandle);
    dlp_ResetSyncFlags(sd, dbHandle);

}
#define DEVICE_USERCONFIG_DB_NAME "MBlnUserConfig"
/* ----------------------------------------------------------------------------*/
static long openUserConfigDatabase(int *threeone)
{
    long result;
    int userConfigDBHandle = 0;

    *threeone = 0;
    result = dlp_OpenDB(sd, DEFAULT_CARD_NUM, dlpOpenRead|dlpOpenWrite,
                        DEVICE_PROFILE_DB_NAME,
                        &userConfigDBHandle);

    if (result < 0) {

        /* OK Now lets look for a 3.1 client database */

        result = dlp_OpenDB(sd, DEFAULT_CARD_NUM, dlpOpenRead|dlpOpenWrite,
                            DEVICE_USERCONFIG_DB_NAME,
                            &userConfigDBHandle);
        if (result < 0) {
            result = dlp_CreateDB(sd, DEVICE_PROFILE_DB_CREATOR,
                                  DEVICE_PROFILE_DB_TYPE, DEFAULT_CARD_NUM,
                                  0, 0, DEVICE_PROFILE_DB_NAME,
                                  &userConfigDBHandle);
            if (result < 0) {
                fprintf(stderr, "Unable to create user Config Databage\n");
                userConfigDBHandle = 0;
            }
        } else {
            *threeone = 1;
        }

    }
    return userConfigDBHandle;
}

#define BUFFERSIZE 0xFFFF
/* ----------------------------------------------------------------------------*/
static int32
readDeviceUserConfig32(int userConfigDBHandle, AGUserConfig **deviceUserConfig)
{
    recordid_t id;
    int bufferSize = BUFFERSIZE;
    int attr = 0;
    int cat  = 0;
    int rc;
    uint8 buffer[BUFFERSIZE];
    AGBufferReader * r = NULL;

#if (PILOT_LINK_MAJOR >= 12)
    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
                               &id, &attr, &cat);
#else
    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
                               &id, &bufferSize, &attr, &cat);
#endif

    if (rc < 0) {
#ifdef _DEBUG
        printf("readDeviceUserConfig: dlp_ReadRecordByIndex , err = %d\n",
               rc);
#endif
        return 0;
    }

    r = AGBufferReaderNew(buffer);
    if (r) {
        *deviceUserConfig = AGUserConfigNew();
        AGUserConfigReadData(*deviceUserConfig, (AGReader*)r);
        AGBufferReaderFree(r);
        return id;
    } else
        return 0;
}
#define BUFFERSIZE 0xFFFF
/*---------------------------------------------------------------------------*/
static int32 readDeviceUserConfig31(int userConfigDBHandle,
                                    AGUserConfig **deviceUserConfig)
{
    recordid_t id;
    int bufferSize = BUFFERSIZE;
    int attr = 0;
    int cat  = 0;
    int rc;
    uint8 buffer[BUFFERSIZE];
    AGBufferReader * r = NULL;

#if (PILOT_LINK_MAJOR >= 12)
    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
                               &id, &attr, &cat);
#else
    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
                               &id, &bufferSize, &attr, &cat);
#endif

    if (rc < 0) {
#ifdef _DEBUG
        printf("readDeviceUserConfig: dlp_ReadRecordByIndex , err = %d\n",
               rc);
#endif
        return 0;
    }

    r = AGBufferReaderNew(buffer);
    if (r) {
        *deviceUserConfig = AGUserConfigNew();
        MAL31ReadUserData(*deviceUserConfig, (AGReader*)r);
        AGBufferReaderFree(r);
        return id;
    } else
        return 0;
}
/*---------------------------------------------------------------------------*/
static int32 readDeviceUserConfig(int userConfigDBHandle,
                                  AGUserConfig **deviceUserConfig,
                                  int threeone)
{
    int32 ret;
    if (threeone)
        ret = readDeviceUserConfig31(userConfigDBHandle, deviceUserConfig);
    else
        ret = readDeviceUserConfig32(userConfigDBHandle, deviceUserConfig);
    return ret;
}

/* ----------------------------------------------------------------------------*/
static void writeDeviceUserConfig(int userConfigDBHandle,
                                  AGUserConfig * deviceUserConfig,
                                  recordid_t *recID, int threeone)
{

    recordid_t id;
    int bufferSize = BUFFERSIZE;
    int attr = 0;
    int cat  = 0;
    uint8 buffer[BUFFERSIZE];
    AGBufferWriter * w = NULL;
    w = AGBufferWriterNew(0);
    if (w) {
        long result;

        if (threeone) {
            MAL31WriteUserData(deviceUserConfig, (AGWriter*)w);
        } else {
            AGUserConfigWriteData(deviceUserConfig, (AGWriter*)w);
        }

#if (PILOT_LINK_MAJOR >= 12)
        result = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
                               &id, &attr, &cat);
#else
        result = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
                               &id, &bufferSize, &attr, &cat);
#endif

        if (result < 0)
            id = 0;

        result =  dlp_WriteRecord(sd, userConfigDBHandle, 0,
                                  id, 0, (void *)AGBufferWriterGetBuffer(w),
                                  AGBufferWriterGetBufferSize(w),
                                  &id);
        AGBufferWriterFree(w);

    }
}

/* ----------------------------------------------------------------------------*/
static AGUserConfig *getUserConfig(uint32 * pilotID)
{

    /* The device is the truth. There is no way to set these values on 
       the desktop */
    AGUserConfig * deviceUserConfig = NULL;
    int userConfigDBHandle = 0;

    /* Get device's record. */
    /*userConfigDBHandle = openUserConfigDatabase();*/
    userConfigDBHandle = openUserConfigDatabase(&threeone);
    if (userConfigDBHandle) {
        
        /* Retrieve device's idea of current userConfig. */
        *pilotID = readDeviceUserConfig(userConfigDBHandle,
                                        &deviceUserConfig,
                                        threeone);

        /* Done with database for now, so close it. */
        dlp_CloseDB(sd, userConfigDBHandle);
        

    }
    return deviceUserConfig;
}

/* ----------------------------------------------------------------------------*/
static void storeDeviceUserConfig(AGUserConfig *userConfig, recordid_t id)
{
    int threeone;
    int userConfigDBHandle = openUserConfigDatabase(&threeone);
    if (0 != userConfigDBHandle) {
        writeDeviceUserConfig(userConfigDBHandle,
                              userConfig, &id, threeone);
        dlp_CloseDB(sd, userConfigDBHandle);
    }
}

/* ----------------------------------------------------------------------------*/
static void doStartServer(PalmSyncInfo * pInfo,
                          AGServerConfig *sc,
                          int32 *errCode)
{
    pInfo->currentServerConfig = sc;
    if(pInfo->commandProcessor) {
        AGCommandProcessorFree(pInfo->commandProcessor);
        pInfo->commandProcessor = NULL;
    }
    pInfo->commandProcessor = AGCommandProcessorNew(sc);
    pInfo->platform->performCommandOut = pInfo->commandProcessor;
    pInfo->platform->performCommandFunc = 
                    AGCommandProcessorGetPerformFunc(pInfo->commandProcessor);

    pInfo->commandProcessor->commands.out = pInfo;

    pInfo->commandProcessor->commands.performTaskFunc = cmdTASK;
    pInfo->commandProcessor->commands.performItemFunc = cmdITEM;
    pInfo->commandProcessor->commands.performDeleteDatabaseFunc = cmdDELETEDATABASE;
    pInfo->commandProcessor->commands.performOpenDatabaseFunc = cmdOPENDATABASE;
    pInfo->commandProcessor->commands.performCloseDatabaseFunc = cmdCLOSEDATABASE;
    pInfo->commandProcessor->commands.performClearModsFunc = cmdCLEARMODS;
    pInfo->commandProcessor->commands.performGoodbyeFunc = cmdGOODBYE;
    pInfo->commandProcessor->commands.performRecordFunc = cmdRECORD;
    
}

/* ----------------------------------------------------------------------------*/
static void doEndServer(PalmSyncInfo * pInfo,
                        int32 *errCode)
{
    pInfo->currentServerConfig = NULL;
    if(pInfo->commandProcessor) {
        AGCommandProcessorFree(pInfo->commandProcessor);
        pInfo->commandProcessor = NULL;
    }
}
/* ----------------------------------------------------------------------------*/
static int32 cmdTASK(void *out, int32 *returnErrorCode, char *currentTask, 
                     AGBool bufferable)
{
    if (currentTask) {
	(*printstatus) ("%s\n", currentTask);
    }
    return AGCLIENT_CONTINUE;
}
/* ----------------------------------------------------------------------------*/
static int32 cmdITEM(void *out, int32 *returnErrorCode,
                                   int32 currentItemNumber,
                                   int32 totalItemCount,
                                   char *currentItem)
{
    (*printstatus) (".");
    if (currentItemNumber == totalItemCount) {
	(*printstatus) ("\n");
    }
    
    return AGCLIENT_CONTINUE;
}
/* ----------------------------------------------------------------------------*/
static int32 cmdDELETEDATABASE(void *out, int32 *returnErrorCode,
                                             char *dbname)
{
    debug_print("doCmdAG_DELETEDATABASE_CMD()\n");

    if (NULL != dbname) {
        debug_print("... trying to delete database ");
        debug_print(dbname);
        debug_print(" from device.\n");
        deleteDatabase(dbname);
    }
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 cmdOPENDATABASE(void *out, int32 *returnErrorCode, char *dbname)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;
    debug_print("doCmdAG_OPENDATABASE_CMD()\n");

    if (NULL != dbname) {
        long result;
        /* We assign a result code here only for debugging.  It's ok for an
        open to fail here. */
        result = openDatabase(pInfo, dbname, TRUE);
    }
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 cmdCLOSEDATABASE(void *out, int32 *returnErrorCode)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;
    debug_print("doCmdAG_CLOSEDATABASE_CMD()\n");
    closeDatabase(pInfo);
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 cmdCLEARMODS(void *out, int32 *returnErrorCode)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;
    debug_print("doCmdAG_CLEARMODS_CMD()\n");
    clearMods(pInfo->pilot_rHandle);
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 cmdGOODBYE(void *out, int32 *returnErrorCode,
                        AGSyncStatus syncStatus,
                        int32 errorCode,
                        char *errorMessage)
{
    debug_print("doCmdAG_GOODBYE_CMD()\n");

    if (errorMessage) {
	(*printerror) ("%s\n", errorMessage);
    }

    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
int getIndexFromPlatformData(uint8 *platformData)
{
    int16 recIndex;
    AGBufferReader reader;

    if (!platformData)
        return 0;

    AGBufferReaderInit(&reader, platformData);
    AGPalmReadRecordPlatformData((AGReader *)&reader, &recIndex);
    AGBufferReaderFinalize(&reader);
    return (int)recIndex;
}
/*----------------------------------------------------------------------------*/
static int32 cmdRECORD(void *out, int32 *returnErrorCode,
                       int32 *newUID,
                       int32 uid,
                       AGRecordStatus mod,
                       int32 recordDataLength,
                       void *recordData,
                       int32 platformDataLength,
                       void *platformData)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;
    
    debug_print("doCmdAG_RECORD_CMD()\n");

    if(mod == AG_RECORD_NEW_TEMPORARY_UID) {
        uid = 0; 
    }

    if (AG_RECORD_DELETED == mod) {

        dlp_DeleteRecord(sd, pInfo->pilot_rHandle, 0, uid);

    } else if (recordDataLength <= 0x0000ffff) {

        if (pInfo->currentDBIsResourceType) {
            
            dlp_WriteRecord(sd, pInfo->pilot_rHandle, dlpDBFlagResource,
                            uid, 0, (void *)recordData,
                            recordDataLength, (recordid_t *)newUID);
            debug_print("doCmdAG_RECORD_CMD()\n");


        } else {

            dlp_WriteRecord(sd, pInfo->pilot_rHandle, 0,
                            uid, 0, (void *)recordData,
                            recordDataLength, (recordid_t *)newUID);
        }

    }
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 initAndOpenDatabase(void *_pInfo, AGDBConfig *db,
                                    int32 *errCode)
{
    long result;
    PalmSyncInfo * pInfo = (PalmSyncInfo *)_pInfo;

    if (NULL == db->dbname) {
        *errCode = AGCLIENT_OPEN_ERR;
        return AGCLIENT_ERR;
    }

    result = openDatabase(pInfo, db->dbname, FALSE);
    if (result < 0) {
        if (result == dlpErrNotFound)
            *errCode = AGCLIENT_OPEN_ERR;
        else
            *errCode = AGCLIENT_UNKNOWN_ERR;
        return AGCLIENT_ERR;
    }

    pInfo->pilot_RecIndex = 0;
    pInfo->record = AGRecordNew(0, AG_RECORD_UNMODIFIED, 0, 0, 0, 0);
    if (!pInfo->record) {
        *errCode = AGCLIENT_OPEN_ERR;
        return AGCLIENT_ERR;
    }

    return AGCLIENT_IDLE;
}
/*----------------------------------------------------------------------------*/
static int32 leaveGetRecord(PalmSyncInfo * pInfo, int32 result)
{
    if (pInfo->record) {

        /* Set recordData to NULL so that AGRecordFree doesn't try
        to free it (we own that buffer and will release it when the
        program ends). */
        pInfo->record->recordData = NULL;

        AGRecordFree(pInfo->record);
        pInfo->record = NULL;
    }
    return result;
}
/*----------------------------------------------------------------------------*/
static int32 getRecordBase(PalmSyncInfo * pInfo, AGBool modonly, 
                           AGRecord ** record, int32 * errCode)
{
    int32 result;
    int att = 0;
    int cat = 0;
    int size = pInfo->pilot_buffer_size;
    int idx   = pInfo->pilot_RecIndex++;

    result = (modonly) ?
#if (PILOT_LINK_MAJOR >= 12)
        dlp_ReadNextModifiedRec (sd, pInfo->pilot_rHandle, pInfo->pilot_buffer,
                                 &pInfo->id, &idx,
                                 &att, &cat)
#else
        dlp_ReadNextModifiedRec (sd, pInfo->pilot_rHandle, pInfo->pilot_buffer,
                                 &pInfo->id, &idx,
                                 &size, &att, &cat)
#endif
        :
#if (PILOT_LINK_MAJOR >= 12)
        dlp_ReadRecordByIndex(sd, pInfo->pilot_rHandle, idx,
                              pInfo->pilot_buffer, &pInfo->id,
                              &att, &cat);
#else
        dlp_ReadRecordByIndex(sd, pInfo->pilot_rHandle, idx,
                              pInfo->pilot_buffer, &pInfo->id,
                              &size, &att, &cat);
#endif
    
    if (result < 0) {
        closeDatabase(pInfo);
        if (result == dlpErrNotFound) {
            debug_print("(successfully reached end of records ...)\n");
            return leaveGetRecord(pInfo, AGCLIENT_IDLE);
        }
        else {
            *errCode = AGCLIENT_UNKNOWN_ERR;
            return leaveGetRecord(pInfo, AGCLIENT_ERR);
        }
    }
    pInfo->record = AGRecordInit(pInfo->record, pInfo->id,
                                 AGPalmPilotAttribsToMALMod((uint8)att),
                                 size, pInfo->pilot_buffer, 0, NULL);

    *record = pInfo->record;
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 getNextModifiedRecord(void * out, AGRecord ** record, 
                                    int32 *errCode)
{
    debug_print("GetNextModifiedRecord()\n");

    return getRecordBase((PalmSyncInfo *)out, TRUE, record, errCode);
}
/*----------------------------------------------------------------------------*/
static int32 getNextRecord(void * out, AGRecord ** record, int32 *errCode)
{
    debug_print("GetNextRecord()\n");

    return getRecordBase((PalmSyncInfo *)out, FALSE, record, errCode);
}
/*----------------------------------------------------------------------------*/
static AGBool doClientProcessorLoop(PalmSyncInfo * pInfo, AGNetCtx *ctx)
{
    int32 dummyError;
    int32 cpResult;
    int32 syncCount;
    int32 i, n;
    AGBool cancelled = FALSE;
    AGLocationConfig *lc = NULL;
    int migrate = 0;

    debug_print("doClientProcessorLoop()\n");

    n = AGUserConfigCount(pInfo->userConfig);

    /* Lets check for 3.1 to 3.x migration here */
    if (n == 1) {
        AGServerConfig * sc =
            AGUserConfigGetServerByIndex(pInfo->userConfig, 0);

        /* Yup, no server name */
        if (!sc->serverName) {
            /* Lets look for the 3.1 config database */
            long result;
            int userConfigDBHandle = 0;
            result = dlp_OpenDB(sd, DEFAULT_CARD_NUM,
                                dlpOpenRead|dlpOpenWrite,
                                DEVICE_USERCONFIG_DB_NAME,
                                &userConfigDBHandle);
            if (result > 0) {
                char response[2];
                printf("It looks like you recently upgraded your client"
                       ". Would you\nlike to migrate your old settings"
                       "?[y/n] ");
                if ((fgets(response, 2, stdin) != 0)
                    && ((response[0] == 'y') || (response[0] == 'Y'))) {

                    threeone = 1;
                    /* Retrieve device's idea of current userConfig. */
                    result = readDeviceUserConfig(userConfigDBHandle,
                                                  &pInfo->userConfig,
                                                  threeone);

                    /* This will cause us to write out the 3.1 data
                     as 3.2 when we finish syncing */
                    threeone = 0;

                    /* Done with database for now, so close it. */
                    dlp_CloseDB(sd, userConfigDBHandle);


                    n = AGUserConfigCount(pInfo->userConfig);

                    /* set a flag to delete the old database */
                    migrate = 1;

                }
            }
        }
    }

    for (i = 0; i < n; ++i) {

        AGServerConfig * sc =
            AGUserConfigGetServerByIndex(pInfo->userConfig, i);
        
        if (cancelled)
            continue;

        if (NULL == sc)
            continue;

        if (sc->disabled)
            continue;

        if (NULL == sc->serverName || sc->serverPort <= 0)
            continue;

        syncCount = 0;
        doStartServer(pInfo, sc, &dummyError);
 
        do {
            AGCommandProcessorStart(pInfo->commandProcessor);
            
            pInfo->deviceInfo = AGDeviceInfoNew();
            /* If this fails, we're totally dead.  Exit with fail code. */
            if(!pInfo->deviceInfo)
                return FALSE;
                
            readDeviceInfo(pInfo);

            if (httpProxy && httpProxyPort) {
                lc = AGLocationConfigNew();
                lc->HTTPUseProxy = 1;
                lc->HTTPName = httpProxy;
                lc->HTTPPort = httpProxyPort;
		(*printstatus) ("Using Proxy Server: Address %s: Port %d\n",
                                httpProxy,
                                httpProxyPort);
            }
            if (proxyUsername && proxyPassword) {
                lc->HTTPUseAuthentication = 1;
                lc->HTTPUsername = proxyUsername;
                lc->HTTPPassword = proxyPassword;
            }

	    (*printstatus) ("Connecting to %s: %s\n", sc->friendlyName, sc->serverName);
            if (socksProxy && socksProxyPort) {
                if (!lc)
                    lc = AGLocationConfigNew();
                lc->SOCKSUseProxy = 1;
                lc->SOCKSName = socksProxy;
                lc->SOCKSPort = socksProxyPort;
		(*printstatus) ("Using SOCKS proxy: Address %s: Port %d\n",
                                socksProxy,
                                socksProxyPort);
            }
            
            pInfo->clientProcessor =
                AGClientProcessorNew(pInfo->currentServerConfig,
                                     pInfo->deviceInfo, 
                                     lc,
                                     pInfo->platform, 
                                     TRUE,
                                     ctx);
                                        
            /* If this fails, we're totally dead.  Exit with fail code. */
            if (NULL == pInfo->clientProcessor) {
                AGDeviceInfoFree(pInfo->deviceInfo);
                return FALSE;
            }
            /*
            if (secnetpresync) {

                if (sc->connectSecurely && sc->allowSecureConnection)
		    (*printstatus) ("Secure Connection\n");

                (*secnetpresync) (ctx, sc, lc,
                                  &pInfo->clientProcessor->syncProcessor,
                                  (sc->connectSecurely &&
                                   sc->allowSecureConnection));
            }
            */
            AGClientProcessorSetBufferServerCommands(pInfo->clientProcessor,
                                                     FALSE);
            
            AGClientProcessorSync(pInfo->clientProcessor);


            cpResult = AGCLIENT_CONTINUE;
            while (AGCLIENT_CONTINUE == cpResult) {
                cpResult = AGClientProcessorProcess(pInfo->clientProcessor);
                
                if (AGCLIENT_CONTINUE == cpResult && pInfo->quit) {
                    cancelled = TRUE;
                    cpResult = AGCLIENT_IDLE;
                }
                if (dlp_OpenConduit(sd)<0) {
                    fprintf(stderr, "Exiting on cancel, data not retrieved.\n");
                    exit(1);
                }
                
            }


            if(cpResult == AGCLIENT_ERR) {

                char *msg = AGGetMsg(pInfo->clientProcessor->errStringId);
                if (msg)
                    /*fprintf(stderr,"%s\n", msg);*/
                    (*printerror) ("%s\n", msg);
                else
                    /*fprintf(stderr,"Unknown error\n");*/
                    (*printerror) ("Unknown error\n");
                    
            }
            /*
            if (secnetpostsync) {
                (*secnetpostsync) (ctx, sc, lc,
                                   &pInfo->clientProcessor->syncProcessor,
                                  (sc->connectSecurely &&
                                   sc->allowSecureConnection));
            }
            */
            AGClientProcessorFree(pInfo->clientProcessor);
            AGDeviceInfoFree(pInfo->deviceInfo);

        } while (!cancelled
                 && AGCommandProcessorShouldSyncAgain(pInfo->commandProcessor) 
                 && syncCount++ < MAX_SYNCS);
        doEndServer(pInfo, &dummyError);

        /* It's possible to cancel in the middle of an open database,
           so if that happened, close it. */
        if (pInfo->pilot_rHandle)
            closeDatabase(pInfo);

        /* If we migrated from 3.1 to 3.2 delete the 3.1 database */
        if (migrate)
            dlp_DeleteDB (sd, DEFAULT_CARD_NUM, DEVICE_USERCONFIG_DB_NAME);

    }

    return TRUE; /* success */
}
/*----------------------------------------------------------------------------*/
void Disconnect(void)
{
    if(sd == 0)
        return;
    
    dlp_EndOfSync(sd, 0);
    pi_close(sd);
    sd = 0;
}
/*----------------------------------------------------------------------------*/
int loadSecLib()
{
    int secureLib = 0;
    void *h;
    char *seclib;

    seclib = getenv("MALSYNC_SECURITYLIB");

    if (!seclib)
        return 0;

#ifdef RTLD_GLOBAL    
    h  = dlopen(seclib, RTLD_GLOBAL|RTLD_NOW);
#else
    h  = dlopen(seclib, RTLD_NOW);
#endif
    if (h) {
        secnetinit        = dlsym(h, "NetInit");
        secnetclose       = dlsym(h, "NetClose");
        secctxsize        = dlsym(h, "NetGetCtxSize");
        secnetpostsync        = dlsym(h, "NetPostSyncHook");
        secnetpresync        = dlsym(h, "NetPreSyncHook");
        
        if (secnetinit && secnetclose && secctxsize ) {
            secureLib = 1; 
        }
    } else {
        /*printf("%s\n", dlerror());*/
        (*printerror) ("%s\n", dlerror());
    }
    
    return secureLib;
}
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
static RETSIGTYPE SigHandler(int signal)
{
  /*fprintf(stderr, "Abort on signal (%d)!", signal);*/
  (*printerror) ("Abort on signal (%d)!", signal);
  Disconnect();
  exit(3);
}
/*----------------------------------------------------------------------------*/
void Connect(PalmSyncInfo *pi)
{
    struct pi_sockaddr addr;
    int ret;
    
    if (sd != 0)
        return;

#ifndef WIN32
    signal(SIGHUP, SigHandler);
    signal(SIGINT, SigHandler);
    signal(SIGSEGV, SigHandler);
#endif

#ifdef PILOT_LINK_0_9
    addr.pi_family = PI_AF_SLP;
    if (!(sd = pi_socket(PI_AF_SLP, PI_SOCK_STREAM, PI_PF_PADP))) {
#else
    addr.pi_family = PI_AF_PILOT;
    if (!(sd = pi_socket(PI_AF_PILOT, PI_SOCK_STREAM, PI_PF_PADP))) {
#endif
        perror("pi_socket");
        exit(1);
    }
    
    strcpy(addr.pi_device, device);
    
#if (PILOT_LINK_MAJOR >= 12)
    ret = pi_bind(sd, (struct sockaddr*)&addr);
#else
    ret = pi_bind(sd, (struct sockaddr*)&addr, sizeof(addr));
#endif
    if(ret == -1) {
        /*fprintf(stderr, "Unable to bind to port '%s'.\n", device);*/
        (*printerror) ("Unable to bind to port '%s'.\n", device);
        exit(1);
    }
    
    if( verbose )
      /*printf("Waiting for connection on %s (press the HotSync button now)...\n",
	     device);*/
	(*printstatus) ("Waiting for connection on %s (press the HotSync button now)...\n", device);
    
    ret = pi_listen(sd,1);
    if(ret == -1) {
        perror("pi_listen");
        exit(1);
    }
    
    sd = pi_accept(sd,0,0);
    if(sd == -1) {
        /*
	if( daemon_mode )
	  return;
        else {
        */
	  perror("pi_accept");
	  exit(1);
	/*}*/
    }
    
    if( verbose )
        /*puts("Connected");*/
	(*printstatus) ("Connected");
}
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/




static int32 getNextCommand(void *out,
                            int32 *newCommand,
                            int32 *commandLength,
                            void **commandBytes)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;
    debug_print("getNextCommand()\n");

    if(!pInfo->sentCommand) {
        AGBufferWriter writer;

        pInfo->sentCommand = TRUE;

        AGBufferWriterInit(&writer, 512);
        AGWriteCompactInt((AGWriter *)&writer, pInfo->subBufferSize);
        AGWriteBytes((AGWriter *)&writer, pInfo->subBuf, pInfo->subBufferSize);

        *newCommand = (int32)AG_XMLDATA_CMD;
        *commandLength = AGBufferWriterGetBufferSize(&writer);
        *commandBytes = AGBufferWriterGetBuffer(&writer);
        writer.buffer = NULL;
        AGBufferWriterFinalize(&writer);

        return AGCLIENT_CONTINUE;

    } else
        return AGCLIENT_IDLE;
}

/*----------------------------------------------------------------------------*/

static char *openSubFile (char **buf, int *bufsize) {
    int fd;
    struct stat statb;
    off_t filesize;
    char *buf2;

    debug_print("openSubFile()\n");

    fd = open (subfile, O_RDONLY);
    if (fd == -1) {
        perror ("open");
        return NULL;
    }
    if (fstat (fd, &statb) == -1) {
        perror ("fstat");
        close (fd);
        return NULL;
    }
    filesize = statb.st_size;
    buf2 = (char *)malloc (filesize + 1);
    if (!buf2) {
        /*fprintf (stderr, "out of memory\n");*/
        (*printerror) ("out of memory\n");
        close (fd);
        return NULL;
    }
    if (read (fd, buf2, filesize) != filesize) {
        perror ("read");
        close (fd);
        return NULL;
    }
    close (fd);
    buf2[filesize] = '\0';

    /* Why +16? */
    *bufsize = strlen(kAvantGoSubXMLWrapper) + filesize + 16;
    *buf = (char*)malloc(*bufsize);

    if (!*buf) {
        /*fprintf (stderr, "Out of memory\n");*/
        (*printerror) ("Out of memory\n");
        free (buf2);
        return NULL;
    }

    sprintf (*buf, kAvantGoSubXMLWrapper, buf2);
    free (buf2);
    return *buf;
}

AGBool submitSubFile (PalmSyncInfo * pInfo, AGNetCtx *ctx)
{
    int32 dummyError;
    int32 cpResult;
    int32 syncCount;
    int32 i, n;
    AGBool cancelled = FALSE;
    AGLocationConfig *lc = NULL;
    uint32 xml_wrapped_data_size;
    char *xml_wrapped_data = NULL;

    debug_print("submitSubFile()\n");

    if (!openSubFile (&xml_wrapped_data, &xml_wrapped_data_size)) {
        return FALSE;
    }

    debug_print("xml data =");
    debug_print(xml_wrapped_data);

    n = AGUserConfigCount(pInfo->userConfig);

    for (i = 0; i < n; ++i) {

        AGServerConfig * sc =
            AGUserConfigGetServerByIndex(pInfo->userConfig, i);

        if (cancelled)
            continue;

        if (NULL == sc)
            continue;

        if (sc->disabled)
            continue;

        if (NULL == sc->serverName || sc->serverPort <= 0)
            continue;

        syncCount = 0;
        doStartServer(pInfo, sc, &dummyError);

        /* Send bytes to server. */

        /* pending(miket) put this somewhere else */
        if (NULL != pInfo->currentServerConfig->serverUri)
            free(pInfo->currentServerConfig->serverUri);
        pInfo->currentServerConfig->serverUri = strdup(kSubmissionURI);
        /* We can't use the nonce that's already stored
         from last sync, so use the data we're sending up
         as a source for the nonce. */
        AGMd5(xml_wrapped_data,
              xml_wrapped_data_size,
              pInfo->currentServerConfig->nonce);

        pInfo->subBuf = xml_wrapped_data;
        pInfo->subBufferSize = xml_wrapped_data_size;
        pInfo->platform->nextExpansionCommandFunc = getNextCommand;

        do {
            AGCommandProcessorStart(pInfo->commandProcessor);

            pInfo->deviceInfo = AGDeviceInfoNew();
            /* If this fails, we're totally dead.  Exit with fail code. */
            if(!pInfo->deviceInfo) {
                free (xml_wrapped_data);
                return FALSE;
            }

            readDeviceInfo(pInfo);

            if (httpProxy && httpProxyPort) {
                lc = AGLocationConfigNew();
                lc->HTTPUseProxy = 1;
                lc->HTTPName = httpProxy;
                lc->HTTPPort = httpProxyPort;
		(*printstatus) ("Using Proxy Server: Address %s: Port %d\n",
                                httpProxy,
                                httpProxyPort);
            }
            if (proxyUsername && proxyPassword) {
                lc->HTTPUseAuthentication = 1;
                lc->HTTPUsername = proxyUsername;
                lc->HTTPPassword = proxyPassword;
            }

	    (*printstatus) ("Connecting to %s: %s\n", sc->friendlyName, sc->serverName);
            if (socksProxy && socksProxyPort) {
                if (!lc)
                    lc = AGLocationConfigNew();
                lc->SOCKSUseProxy = 1;
                lc->SOCKSName = socksProxy;
                lc->SOCKSPort = socksProxyPort;
		(*printstatus) ("Using SOCKS proxy: Address %s: Port %d\n",
                                socksProxy,
                                socksProxyPort);
            }

            pInfo->clientProcessor =
                AGClientProcessorNew(pInfo->currentServerConfig,
                                     pInfo->deviceInfo,
                                     lc,
                                     pInfo->platform,
                                     TRUE,
                                     ctx);

            /* If this fails, we're totally dead.  Exit with fail code. */
            if (NULL == pInfo->clientProcessor) {
                AGDeviceInfoFree(pInfo->deviceInfo);
                free (xml_wrapped_data);
                return FALSE;
            }

            if (secnetpresync) {

                if (sc->connectSecurely && sc->allowSecureConnection)
		    (*printstatus) ("Secure Connection\n");

                (*secnetpresync) (ctx, sc, lc,
                                  &pInfo->clientProcessor->syncProcessor,
                                  (sc->connectSecurely &&
                                   sc->allowSecureConnection));
            }

            AGClientProcessorSetBufferServerCommands(pInfo->clientProcessor,
                                                     TRUE);

            AGClientProcessorSync(pInfo->clientProcessor);


            cpResult = AGCLIENT_CONTINUE;
            while (AGCLIENT_CONTINUE == cpResult) {
                cpResult = AGClientProcessorProcess(pInfo->clientProcessor);

                if (AGCLIENT_CONTINUE == cpResult && pInfo->quit) {
                    cancelled = TRUE;
                    cpResult = AGCLIENT_IDLE;
                }
                if (dlp_OpenConduit(sd)<0) {
		    (*printstatus) ("Exiting on cancel, data not retrieved.\n");
                    exit(1);
                }

            }


            if(cpResult == AGCLIENT_ERR) {

                char *msg = AGGetMsg(pInfo->clientProcessor->errStringId);
                if (msg)
                    /*fprintf(stderr,"%s\n", msg);*/
                    (*printerror) ("%s\n", msg);
                else
                    /*fprintf(stderr,"Unknown error\n");*/
                    (*printerror) ("Unknown error\n");

            }

            if (secnetpostsync) {
                (*secnetpostsync) (ctx, sc, lc,
                                   &pInfo->clientProcessor->syncProcessor,
                                   (sc->connectSecurely &&
                                    sc->allowSecureConnection));
            }

            AGClientProcessorFree(pInfo->clientProcessor);
            AGDeviceInfoFree(pInfo->deviceInfo);

        } while (!cancelled
                 && AGCommandProcessorShouldSyncAgain(pInfo->commandProcessor)
                 && syncCount++ < 5);
        doEndServer(pInfo, &dummyError);

        /* It's possible to cancel in the middle of an open database,
         so if that happened, close it. */
        if (pInfo->pilot_rHandle)
            closeDatabase(pInfo);

    }

    free (xml_wrapped_data);
    return TRUE; /* success */
}

int malsync (int sd_hack, PalmSyncInfo *pInfo) {
    uint32 pilotID;
    AGNetCtx *ctx;

    sd = sd_hack;

    if (loadSecLib()) {
        ctx = calloc(1, (*secctxsize)());
        (*secnetinit)(ctx);
    } else {
        ctx = calloc( 1, sizeof( AGNetCtx));
        AGNetInit( ctx);
    }

    if (setupPlatformCalls(pInfo))
        return -1;

    pInfo->userConfig = getUserConfig(&pilotID);
    if (pInfo->userConfig) {
        doClientProcessorLoop (pInfo, ctx);
        storeDeviceUserConfig (pInfo->userConfig, pilotID);
    }
    else {
        (*printerror) ("**** ERROR: No SyncMAL user config found!\n");
        (*printerror) ("Is AvantGo installed and configured on the PDA?\n");
    }

    if (secnetclose) {
        (*secnetclose)(ctx);
    } else {
        AGNetClose(ctx);
    }

    free(ctx);

    return 0;
}
